/*
  M-convexity test
  via Hessian matrix on all points inside domain
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"

/* INPUT  func : pointer to the function */
/* INPUT  x    : test point in domain */
/* WORK   a    : working matrix */
static int mv(int dim, double func(int,int*), int x[], double **a) {
    make_hesse_m(dim, func, x, a);
    return is_mconv_quadratic(dim, a);
}

static int mv2(int dim, double func(int,int*), int x[], double **a) {
    make_hesse_m2(dim, func, x, a);
    return is_mgconv_quadratic(dim-1, a);  /* one dimension smaller and of Mg */
}


/* loop for high dimension (only for M-convex): usage
   do {
     ...
   } while (ret && next(dim, x, lower, upper, step));
*/
static int nextm(int dim, int x[], int lower[], int upper[], int step) {
    int i = 0;

    for (i=0; i<dim-1; i++) {
        x[i] += step;  x[dim-1] -= step;
        if (x[i] <= upper[i] && x[dim-1] >= lower[dim-1]) return TRUE;
        x[dim-1] += x[i]-lower[i];
        x[i] = lower[i];
    }
    return FALSE; /* end mark */
}


/* decide if the function is M-convex or not */
static int is_convex(int dim, double func(int,int*),
                     int lower[], int upper[], int sum, const int type) {
    int ret = TRUE, step;
    int *x = ivector_new(dim);
    double **a = fmatrix_new(dim, dim);
    int i;

    ivector_copy(dim, x, lower);

    for (i=0; i<dim-1; i++) sum -= x[i];
    x[dim-1] = sum;
    if (upper[dim-1]<x[dim-1]) nextm(dim, x, lower, upper, 1);

    step = upper[0]-lower[0];
    while ((step & (step-1)) != 0) step &= step - 1;

    for (step--; ret && step>=1; step/=2) {
        do {
#ifdef DEBUG
            printf("x=");ivector_print(dim, x);
            printf("\n");
#endif /* DEBUG */
            if   (type == 1) ret =  mv(dim, func, x, a);
            else             ret = mv2(dim, func, x, a);
            if (verbose_level & STATUS && !ret) {
                printf("x=");    ivector_print(dim, x);
                printf(", step=%d\n", step);
                printf("hesse=");fmatrix_print(dim, a);
                printf("\n");
            }
        } while (ret && nextm(dim, x, lower, upper, step));
    }

    free(a);
    free(x);
    return ret;
}


/* decide if the function is M-convex or not */
int  is_mconv_gene(int dim, double func(int,int*), int lower[], int upper[],
                   int sum) {
    return is_convex(dim, func, lower, upper, sum, 1);
}

/* decide if the function is M-convex or not, another implementation */
int  is_mconv_gene2(int dim, double func(int,int*), int lower[], int upper[],
                    int sum) {
    return is_convex(dim, func, lower, upper, sum, 2);
}
